# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EventProvider identification flow:

       Create
      unknown
       event
         |
         V
    Is APIGateway? --(yes)--> Read Body, Parameters
         |                             |
        (no)                         (then)
         |-----------------------------|
         V
     Read EventProvider.
     Storage keys? --(yes)-->  Read keys (minio, s3, onedata)
         |                             |
        (no)                           |
         |-----------------------------|
         V
    Return parsed
       event
"""

import json
from faassupervisor.events.apigateway import ApiGatewayEvent
from faassupervisor.events.minio import MinioEvent
from faassupervisor.events.onedata import OnedataEvent
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.dCache import DCacheEvent
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.logger import get_logger
from faassupervisor.exceptions import exception, UnknowStorageEventWarning
from faassupervisor.utils import SysUtils

_S3_EVENT = "aws:s3"
_MINIO_EVENT = "minio:s3"
_ONEDATA_EVENT = "OneTrigger"
_DCACHE_EVENT = "dcacheTrigger"

def _is_api_gateway_event(event_info):
    return 'httpMethod' in event_info


def _is_storage_event(event_info):
    if 'Records' in event_info \
           and event_info['Records'] \
           and 'eventSource' in event_info['Records'][0]:
        return event_info['Records'][0]['eventSource'] == _S3_EVENT \
            or event_info['Records'][0]['eventSource'] == _MINIO_EVENT \
            or event_info['Records'][0]['eventSource'] == _ONEDATA_EVENT \
            or event_info['Records'][0]['eventSource'] == _DCACHE_EVENT
    return False


def _is_delegated_event(event_info):
    return 'event' in event_info


@exception()
def _parse_storage_event(event, storage_provider='default'):
    record = event['Records'][0]['eventSource']
    if record == _S3_EVENT:
        parsed_event = S3Event(event, storage_provider)
        get_logger().info("S3 event created")
    elif record == _MINIO_EVENT:
        parsed_event = MinioEvent(event, storage_provider)
        get_logger().info("MINIO event created")
    elif record == _ONEDATA_EVENT:
        parsed_event = OnedataEvent(event)
        get_logger().info("ONEDATA event created")
    elif record == _DCACHE_EVENT:
        parsed_event = DCacheEvent(event, storage_provider)
        get_logger().info("DCACHE event created")
    else:
        raise UnknowStorageEventWarning()
    return parsed_event


def _set_storage_env_vars(parsed_event, event):
    # Store 'object_key' in environment variable
    SysUtils.set_env_var("STORAGE_OBJECT_KEY", parsed_event.object_key)
    # Store 'event_time' in environment variable
    SysUtils.set_env_var("EVENT_TIME", parsed_event.event_time)
    # Store the raw event in environment variable
    SysUtils.set_env_var("EVENT", json.dumps(event))


def parse_event(event, storage_provider="default"):
    """Parses the received event and
    returns the appropriate event class."""
    # Make sure the event is always stored
    parsed_event = None
    if not isinstance(event, dict):
        try:
            event = json.loads(event)
        except ValueError:
            return UnknownEvent(event)
    # Applies the event identification flow
    if _is_api_gateway_event(event):
        get_logger().info("API Gateway event found.")
        parsed_event = ApiGatewayEvent(event)
        return parse_event(parsed_event.body)
    if _is_delegated_event(event):
        get_logger().info("Delegated event found.")
        if 'storage_provider' in event:
            return parse_event(event['event'], event['storage_provider'])
        return parse_event(event['event'])
    if _is_storage_event(event):
        get_logger().info("Storage event found.")
        parsed_event = _parse_storage_event(event, storage_provider)
        _set_storage_env_vars(parsed_event, event)
    return parsed_event if parsed_event else UnknownEvent(event)



from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.utils import FileUtils

class DCacheEvent(UnknownEvent):
    
    _TYPE = 'DCACHE'
    
    def __init__(self, event, provider_id='default'):
        super().__init__(event.get('event') or event)
        self.provider_id = provider_id

    def _set_event_params(self):
        self.object_key = self.event["file_path"]
        self.file_name = FileUtils.get_file_name(self.object_key)
        self.event_time = self.event["timestamp"]


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In this module are defined the classes and methods
used to manage the batch jobs from the lambda environment."""

import json
import boto3
from faassupervisor.utils import ConfigUtils


class Batch():
    """Used for defining Batch jobs in the Batch environment."""

    # pylint: disable=too-few-public-methods

    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance
        self.script = self._get_user_script()
        self._create_context()
        self._set_job_variables()
        self._create_batch_client()

    def _create_context(self):
        self.context = {
            'function_name': self.lambda_instance.get_function_name(),
            'memory_limit_in_mb': self.lambda_instance.get_memory(),
            'aws_request_id': self.lambda_instance.get_request_id(),
            'log_group_name': self.lambda_instance.get_log_group_name(),
            'log_stream_name': self.lambda_instance.get_log_stream_name()
        }

    def _create_batch_client(self):
        self.client = boto3.client('batch')

    def _set_job_variables(self):
        self.batch_job_env_vars = []
        if self.script:
            self._add_batch_job_env_var("SCRIPT", self.script)
        self._add_batch_job_env_var("EVENT", json.dumps(self.lambda_instance.raw_event))
        self._add_batch_job_env_var("CONTEXT", json.dumps(self.context))
        self._add_batch_job_env_var("AWS_LAMBDA_REQUEST_ID", self.context.get("aws_request_id"))

    def _add_batch_job_env_var(self, name, value):
        if name and value:
            self.batch_job_env_vars.append({"name": name, "value": value})

    def _get_user_script(self):
        script = ''
        if 'script' in self.lambda_instance.raw_event:
            script = self.lambda_instance.raw_event['script']
        return script

    def _get_overrides(self):
        batch = ConfigUtils.read_cfg_var("batch")
        if batch.get("multi_node_parallel").get("enabled") is True:
            return {
                "nodeOverrides": {
                    "nodePropertyOverrides": [
                        {
                            "containerOverrides": {
                                "environment": self.batch_job_env_vars
                            },
                            "targetNodes": "0:"
                        }
                    ]
                }
            }
        else:
            return {
                "containerOverrides": {
                    "environment": self.batch_job_env_vars
                }
            }

    def _get_job_args(self):
        job_name = self.context.get("function_name")
        overrides = self._get_overrides()
        job_def = {
            "jobDefinition": job_name,
            "jobName": job_name,
            "jobQueue": job_name,
        }
        return {**job_def, **overrides}

    def _submit_batch_job(self, job_args):
        return self.client.submit_job(**job_args)["jobId"]

    def invoke_batch_function(self):
        """Submit job from a lambda instance."""
        return self._submit_batch_job(self._get_job_args())



from faassupervisor.storage.providers import DefaultStorageProvider
from faassupervisor.utils import SysUtils, FileUtils
from webdav3.client import Client


class WebDav(DefaultStorageProvider):
    _TYPE = "WEBDAV"

    def __init__(self, stg_auth):
        super().__init__(stg_auth)
        self.client = self._get_client()

    def _get_client(self):
        """Returns a WebDav client to connect to the https endpoint of the storage provider"""
        options = {
            'webdav_hostname': 'https://'+self.stg_auth.get_credential('hostname'),
            'webdav_login':    self.stg_auth.get_credential('login'),
            'webdav_password': self.stg_auth.get_credential('password')
        }
        return Client(options=options)

    def download_file(self, parsed_event, input_dir_path):
        file_download_path = SysUtils.join_paths(input_dir_path, parsed_event.file_name)
        self.client.download_sync(remote_path=parsed_event.object_key, local_path=file_download_path)

    def upload_file(self, file_path, file_name, output_path):
        if self.client.check(output_path):
            self.client.upload_sync(remote_path=output_path+"/"+file_name, local_path=file_path)
        else:
            self.client.mkdir(output_path)
            self.client.upload_sync(remote_path=output_path+"/"+file_name, local_path=file_path)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In this module are defined the classes and methods
used to manage the container runtime jobs from the lambda environment."""

import subprocess
import os.path
import os

from faassupervisor.utils import FileUtils, SysUtils
from faassupervisor.logger import get_logger
from faassupervisor.exceptions import ContainerTimeoutExpiredWarning


class Container():
    """Used for defining Container jobs in the Lambda container runtime."""

    _CONTAINER_OUTPUT_FILE = SysUtils.join_paths(FileUtils.get_tmp_dir(), "container-stdout")

    def __init__(self, lambda_instance):
        self.lambda_instance = lambda_instance
        self.script = None

        if hasattr(self.lambda_instance, 'script_path'):
            self.script = self.lambda_instance.script_path
        # Script to be executed every time (if defined)
        elif hasattr(self.lambda_instance, 'init_script_path'):
            self.script = self.lambda_instance.init_script_path
        else:
            raise Exception("Init: Script not defined.")
        if not os.path.isfile(self.script):
            raise Exception("Init: Script %s does not exist." % self.script)

    def invoke_function(self):
        if self.script:
            remaining_seconds = self.lambda_instance.get_remaining_time_in_seconds()
            get_logger().debug("Executing command: %s" % self.script)

            new_env = os.environ.copy()
            new_env.update(SysUtils.get_cont_env_vars())
            # Remove the library path set by Pyinstaller
            if 'LD_LIBRARY_PATH_ORIG' in new_env:
                new_env['LD_LIBRARY_PATH'] = new_env['LD_LIBRARY_PATH']
            elif 'LD_LIBRARY_PATH' in new_env:
                del new_env['LD_LIBRARY_PATH']

            with open(self._CONTAINER_OUTPUT_FILE, "wb") as out:
                with subprocess.Popen(['/bin/sh', self.script],
                                      stderr=subprocess.STDOUT,
                                      stdout=out,
                                      env=new_env,
                                      start_new_session=True) as process:
                    try:
                        rc = process.wait(timeout=remaining_seconds)
                        if rc != 0:
                            get_logger().warning("User script exited with code %s!" % rc)
                    except subprocess.TimeoutExpired:
                        get_logger().info("Stopping process '%s'", process)
                        process.kill()
                        raise ContainerTimeoutExpiredWarning()

            output = b''
            if FileUtils.is_file(self._CONTAINER_OUTPUT_FILE):
                output = FileUtils.read_file(self._CONTAINER_OUTPUT_FILE, file_mode="rb")
            return output
        else:
            return b''



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods shared by all the classes."""

import base64
import json
import os
import subprocess
import sys
import tempfile
import shutil
import yaml
from zipfile import ZipFile


class SysUtils():
    """Common methods for system management."""

    @staticmethod
    def get_stdin():
        """Returns system stdin."""
        return sys.stdin.read()

    @staticmethod
    def join_paths(*paths):
        """Returns the strings passed joined as a system path."""
        return os.path.join(*paths)

    @staticmethod
    def is_var_in_env(variable):
        """Checks if a variable is in the system environment."""
        return variable in os.environ

    @staticmethod
    def set_env_var(key, value):
        """Sets a system environment variable."""
        if key and value:
            os.environ[key] = value

    @staticmethod
    def get_all_env_vars():
        """Returns all the system environment variables."""
        return os.environ

    @staticmethod
    def get_env_var(variable):
        """Returns the value of system environment variable
        or an empty string if not found."""
        return os.environ.get(variable, "")

    @staticmethod
    def delete_env_var(variable):
        """Deletes the specified variable from the environment."""
        if variable in os.environ:
            os.environ.pop(variable)

    @staticmethod
    def get_cont_env_vars():
        """Returns the defined container environment variables."""
        container = ConfigUtils.read_cfg_var('container')
        return container.get('environment', {}).get('Variables', {})

    @staticmethod
    def execute_cmd(command):
        """Executes a bash command."""
        subprocess.call(command)

    @staticmethod
    def execute_cmd_and_return_output(command, encoding='utf-8'):
        """Executes a bash command and returns the console output."""
        return subprocess.check_output(command).decode(encoding)

    @staticmethod
    def is_lambda_environment():
        """Checks if supervisor is running in AWS Lambda."""
        return (SysUtils.is_var_in_env('AWS_EXECUTION_ENV') and
                SysUtils.get_env_var('AWS_EXECUTION_ENV').startswith('AWS_Lambda_'))

    @staticmethod
    def is_lambda_image_environment():
        """Checks if supervisor is running in AWS Lambda Imagen runtime."""
        return (SysUtils.is_var_in_env('AWS_EXECUTION_ENV') and
                SysUtils.get_env_var('AWS_EXECUTION_ENV') == 'AWS_Lambda_Image')


class FileUtils():
    """Common methods for file and directory management."""

    @staticmethod
    def set_file_execution_rights(file_path):
        """Makes a file executable."""
        # Execution rights for user, group and others
        mode = os.stat(file_path).st_mode | 0o0111
        os.chmod(file_path, mode)

    @staticmethod
    def cp_file(file_src, file_dst):
        """Copy file to specified destination."""
        shutil.copyfile(file_src, file_dst)

    @staticmethod
    def create_folder(folder_name):
        """Creates a system folder.
        Does nothing if the folder exists."""
        os.makedirs(folder_name, exist_ok=True)

    @staticmethod
    def create_file_with_content(path, content, mode='w'):
        """Creates a new file with the passed content.
        If the content is a dictionary, first is converted to a string."""
        with open(path, mode) as fwc:
            if isinstance(content, dict):
                content = json.dumps(content)
            fwc.write(content)

    @staticmethod
    def read_file(file_path, file_mode="r", file_encoding="utf-8"):
        """Reads the whole specified file and returns the content."""
        content = ''
        if file_mode == 'rb':
            file_encoding = None
        with open(file_path, mode=file_mode, encoding=file_encoding) as content_file:
            content = content_file.read()
        return content

    @staticmethod
    def create_tmp_dir():
        """Creates a directory in the temporal folder of the system.
        When the context is finished, the folder is automatically deleted."""
        return tempfile.TemporaryDirectory()

    @staticmethod
    def get_tmp_dir():
        """Gets the directory where the temporal
        folder of the system is located."""
        return tempfile.gettempdir()

    @staticmethod
    def get_all_files_in_dir(dir_path):
        """Returns a list with all the file paths in
        the specified directory and subdirectories."""
        files = []
        for dirpath, _, filenames in os.walk(dir_path):
            for filename in filenames:
                files.append(SysUtils.join_paths(dirpath, filename))
        return files

    @staticmethod
    def is_file(file_path):
        """Test whether a path is a regular file."""
        return os.path.isfile(file_path)

    @staticmethod
    def get_file_name(file_path):
        """Returns the filename."""
        return os.path.basename(file_path)

    @staticmethod
    def get_dir_name(file_path):
        """Returns the directory name."""
        return os.path.dirname(file_path)

    @staticmethod
    def zip_file_list(file_list, zip_path):
        """Generate a zip in zip_path from a list of files"""
        dir_name = os.path.dirname(zip_path)
        with ZipFile(zip_path, 'w') as zip_file:
            for file in file_list:
                zip_file.write(file, StrUtils.remove_prefix(file, '{}/'.format(dir_name)))


class StrUtils():
    """Common methods for string management."""

    @staticmethod
    def bytes_to_base64str(value, encoding='utf-8'):
        """Encodes string to base64 and returns another string."""
        return base64.b64encode(value).decode(encoding)

    @staticmethod
    def dict_to_base64str(value, encoding='utf-8'):
        """Encodes a dictionary to base64 and returns a string."""
        return base64.b64encode(json.dumps(value).encode(encoding)).decode(encoding)

    @staticmethod
    def base64_to_str(value, encoding='utf-8'):
        """Decodes from base64 and returns a string."""
        return base64.b64decode(value).decode(encoding)

    @staticmethod
    def utf8_to_base64_string(value):
        """Encode a 'utf-8' string using Base64 and return
        the encoded value as a string."""
        return StrUtils.bytes_to_base64str(value.encode('utf-8'))

    @staticmethod
    def get_storage_id(io_storage):
        """Return the identifier of the storage provider or 'default'
        if it is not set.
        Format: <PROVIDER_TYPE>.<ID>"""
        if isinstance(io_storage, str):
            split = io_storage.split('.', maxsplit=1)
            if len(split) == 2:
                return split[1]
        return 'default'

    @staticmethod
    def get_storage_type(io_storage):
        """Return the provider type of the storage provider.
        Format: <PROVIDER_TYPE>.<ID>"""
        if isinstance(io_storage, str):
            split = io_storage.split('.', maxsplit=1)
            return split[0].upper()
        return None

    @staticmethod
    def remove_prefix(text, prefix):
        """Return a substring without the specified prefix
        or the same text if the prefix is not found."""
        if text.startswith(prefix):
            text = text.replace(prefix, '', 1)
        return text


class ConfigUtils():
    """Common methods for configuration file management."""

    _LAMBDA_STORAGE_CONFIG_PATH = '/var/task/function_config.yaml'
    _LAMBDA_STORAGE_CONFIG_ENV = 'FDL'
    _BINARY_STORAGE_CONFIG_ENV = 'FUNCTION_CONFIG'
    _BINARY_OSCAR_STORAGE_CONFIG_PATH = '/oscar/config/function_config.yaml'
    _CUSTOM_VARIABLES = [
        'log_level',
        'execution_mode',
        'extra_payload',
        'udocker_exec',
        'udocker_dir',
        'udocker_bin',
        'udocker_lib',
        'download_input'
    ]

    @classmethod
    def read_cfg_var(cls, variable):
        """Returns the value of a config variable or an empty
        string if not found."""
        if SysUtils.is_lambda_environment():
            # Check if config file exsits in '_LAMBDA_STORAGE_CONFIG_PATH'
            if FileUtils.is_file(cls._LAMBDA_STORAGE_CONFIG_PATH):
                # Read config file
                with open(cls._LAMBDA_STORAGE_CONFIG_PATH) as file:
                    config = yaml.safe_load(file)
            else:
                # Get and decode content of '_LAMBDA_STORAGE_CONFIG_ENV'
                encoded = SysUtils.get_env_var(cls._LAMBDA_STORAGE_CONFIG_ENV)
                decoded = StrUtils.base64_to_str(encoded)
                config = yaml.safe_load(decoded)
        else:
            # Check if config file exsits in '_BINARY_OSCAR_STORAGE_CONFIG_PATH'
            if FileUtils.is_file(cls._BINARY_OSCAR_STORAGE_CONFIG_PATH):
                # Read config file
                with open(cls._BINARY_OSCAR_STORAGE_CONFIG_PATH) as file:
                    config = yaml.safe_load(file)
            else:
                # Get and decode content of '_BINARY_STORAGE_CONFIG_ENV'
                encoded = SysUtils.get_env_var(cls._BINARY_STORAGE_CONFIG_ENV)
                decoded = StrUtils.base64_to_str(encoded)
                config = yaml.safe_load(decoded)
        # Manage variables that could be defined in environment
        if variable in cls._CUSTOM_VARIABLES:
            value = SysUtils.get_env_var(variable.upper())
            if value != '':
                return value
        return config.get(variable, '') if config else ''



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.faas module and classes."""

import unittest
from unittest import mock
import os
import subprocess
# from faassupervisor.events.minio import MinioEvent
from faassupervisor.faas.binary.supervisor import BinarySupervisor
from faassupervisor.faas.aws_lambda.supervisor import LambdaSupervisor, \
                                                      is_batch_execution, \
                                                      _is_lambda_batch_execution
from faassupervisor.exceptions import NoLambdaContextError
# from faassupervisor.storage.config import StorageConfig
# from faassupervisor.supervisor import Supervisor
# from faassupervisor.utils import FileUtils, StrUtils
# from faassupervisor.utils import ConfigUtils

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access

MINIO_EVENT = {"Key": "images/nature-wallpaper-229.jpg",
               "Records": [{"s3": {"object": {"key": "nature-wallpaper-229.jpg"},
                                   "bucket": {"name": "images",
                                              "arn": "arn:aws:s3:::images"}},
                            "eventSource": "minio:s3",
                            "eventTime": "2018-06-29T10:23:44Z"}]}

DELEGATED_MINIO_EVENT = {"storage_provider": "minio.cluster2",
                         "event": MINIO_EVENT
                        }
CONFIG_FILE_OK = """
name: test-func
input:
- storage_provider: minio.test_minio
  path: files
output:
- storage_provider: s3
  path: bucket/folder
- storage_provider: minio.test_minio
  path: bucket
storage_providers:
  minio:
    test_minio:
        access_key: test_minio_access
        secret_key: test_minio_secret    
    cluster2:
        access_key: test_minio_access
        secret_key: test_minio_secret
"""

class BinarySupervisorTest(unittest.TestCase):

    @mock.patch('subprocess.Popen')
    @mock.patch('faassupervisor.utils.FileUtils.create_file_with_content')
    def test_execute_function(self, mock_create, mock_popen):
        supervisor = BinarySupervisor('UNKNOWN')
        with mock.patch.dict('os.environ', {'SCRIPT': 'ZmFrZSBzY3JpcHQh',
                                            'TMP_INPUT_DIR': '/tmp/input'}, clear=True):
            supervisor.execute_function()
            # Check script file creation
            mock_create.assert_called_once_with('/tmp/input/script.sh', 'fake script!')
            # Check process execution
            mock_popen.assert_called_once_with(['/bin/sh', '/tmp/input/script.sh'],
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.STDOUT,
                                               encoding='utf-8',
                                               errors='ignore')


class LambdaSupervisorTest(unittest.TestCase):

    def _get_context(self):
        mock_context = mock.Mock()
        type(mock_context).aws_request_id = mock.PropertyMock(return_value='123')
        mock_context.get_remaining_time_in_millis.return_value = 123
        return mock_context

    def test_is_batch_execution(self):
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'batch'}, clear=True):
            self.assertTrue(is_batch_execution())
        with mock.patch.dict('os.environ', {}, clear=True):
            self.assertFalse(is_batch_execution())

    def test_is_lambda_batch_execution(self):
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'lambda-batch'}, clear=True):
            self.assertTrue(_is_lambda_batch_execution())
        with mock.patch.dict('os.environ', {}, clear=True):
            self.assertFalse(_is_lambda_batch_execution())

    def test_create_lambda_supervisor(self):
        with self.assertRaises(NoLambdaContextError):
            LambdaSupervisor(None, None)
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'lambda-batch',
                                            'TMP_INPUT_DIR': '/tmp/input',
                                            'TMP_OUTPUT_DIR': '/tmp/output'}, clear=True):

            LambdaSupervisor('event', self._get_context())

    @mock.patch('subprocess.Popen')
    @mock.patch('faassupervisor.utils.FileUtils.cp_file')
    @mock.patch('faassupervisor.utils.SysUtils.execute_cmd')
    @mock.patch('faassupervisor.utils.SysUtils.execute_cmd_and_return_output')
    @mock.patch('faassupervisor.faas.aws_lambda.udocker.get_function_ip')
    @mock.patch('faassupervisor.utils.ConfigUtils.read_cfg_var')
    def test_execute_function(self, mock_read_cfg_var, mock_get_function_ip, mock_execute_out,
                              mock_execute, mock_cp_file, mock_popen):
        mock_read_cfg_var.side_effect = ["1", "init_script.sh", "3", {"image": "image"},
                                         {"image": "image"}, {"timeout_threshold": 10}]
        mock_execute_out.return_value = "22"
        mock_get_function_ip.return_value = "127.0.0.1"
        with mock.patch.dict('os.environ', {'EXECUTION_MODE': 'lambda-batch',
                                            'TMP_INPUT_DIR': '/tmp/input',
                                            'UDOCKER_DIR': '/tmp/udocker',
                                            'UDOCKER_EXEC': '/udocker.py',
                                            'TMP_OUTPUT_DIR': '/tmp/output'}, clear=True):
            supervisor = LambdaSupervisor('event', self._get_context())
            supervisor.execute_function()

        res = ['/udocker.py', '--quiet', 'run', '-v', '/tmp/input', '-v', '/tmp/output', '-v',
               '/dev', '-v', '/proc', '-v', '/etc/hosts', '--nosysdirs', '--env', 'REQUEST_ID=123',
               '--env', 'INSTANCE_IP=127.0.0.1', '--env', 'TMP_OUTPUT_DIR=/tmp/output',
               '--entrypoint=/bin/sh /tmp/input/init_script.sh', 'udocker_container']
        self.assertEqual(mock_popen.call_args_list[0][0][0], res)

    @mock.patch('subprocess.Popen')
    @mock.patch('os.path.isfile')
    @mock.patch('faassupervisor.utils.ConfigUtils.read_cfg_var')
    @mock.patch('faassupervisor.utils.FileUtils.cp_file')
    def test_execute_function_container(self, mock_cp_file, mock_read_cfg_var, mock_is_file, mock_popen):
        mock_read_cfg_var.side_effect = ["1", "init_script.sh", {"timeout_threshold": 10}, {}]
        with mock.patch.dict('os.environ', {'AWS_EXECUTION_ENV': 'AWS_Lambda_Image',
                                            'TMP_INPUT_DIR': '/tmp/input',
                                            'TMP_OUTPUT_DIR': '/tmp/output'}, clear=True):
            supervisor = LambdaSupervisor('event', self._get_context())
            supervisor.execute_function()

        res = ['/bin/sh', '/tmp/input/init_script.sh']
        self.assertEqual(mock_popen.call_args_list[0][0][0], res)
        res =  {'AWS_EXECUTION_ENV': 'AWS_Lambda_Image',
                'TMP_INPUT_DIR': '/tmp/input',
                'TMP_OUTPUT_DIR': '/tmp/output',
                'AWS_LAMBDA_REQUEST_ID': '123'}
        self.assertEqual(mock_popen.call_args_list[0][1]['env'], res)

# class SupervisorTest(unittest.TestCase):
#     @mock.patch('faassupervisor.utils.ConfigUtils.read_cfg_var')
#     @mock.patch('faassupervisor.supervisor._create_supervisor')
#     @mock.patch('faassupervisor.supervisor.Supervisor._parse_input')
#     @mock.patch('faassupervisor.storage.config.StorageConfig._parse_config')
#     @mock.patch('faassupervisor.storage.config.StorageConfig.download_input')
#     def test_parse_input(self, mock_download_input, mock_parse_config , mock_parse_input, mock_create_supervisor, mock_read_cfg_var):
#             with mock.patch.dict('os.environ', {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK),
#                                             'DOWNLOAD_INPUT': ''}, clear=True):

#                 supervisor = Supervisor(DELEGATED_MINIO_EVENT)
#                 parsed_event = MinioEvent(MINIO_EVENT, provider_id = 'cluster2')
#                 supervisor._parse_input()
#                 mock_create_supervisor.assert_called_with(DELEGATED_MINIO_EVENT, None, 'MINIO')
#                 mock_parse_config.assert_called()
#                 mock_read_cfg_var.assert_called_with('output')
#                 mock_parse_input.assert_called()
#                 self.assertEqual(supervisor.parsed_event.get_type(), parsed_event.get_type())
#                 mock_download_input.assert_called_with(parsed_event, '/tmp/input')
#                 #mock_read_cfg_var.assert_called_with('storage_providers')
# """             minio_auth = StorageConfig()._get_auth_data('MINIO', 'cluster2')
#                 self.assertEqual(minio_auth['test_minio'].type, 'MINIO') """



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class to parse, store and manage storage information."""

from faassupervisor.storage.providers.webdav import WebDav
from faassupervisor.utils import ConfigUtils, FileUtils, StrUtils
from faassupervisor.exceptions import StorageAuthError, \
    InvalidStorageProviderError, exception
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.s3 import S3
from faassupervisor.logger import get_logger


def create_provider(storage_auth):
    """Returns the storage provider needed
    based on the authentication type defined.

    If not storage auth provided, use local storage."""
    if not storage_auth:
        provider = Local(storage_auth)
    elif storage_auth.type == 'MINIO':
        provider = Minio(storage_auth)
    elif storage_auth.type == 'ONEDATA':
        provider = Onedata(storage_auth)
    elif storage_auth.type == 'S3':
        provider = S3(storage_auth)
    elif storage_auth.type == 'WEBDAV':
        provider = WebDav(storage_auth)
    else:
        raise InvalidStorageProviderError(storage_type=storage_auth.type)
    return provider


class AuthData():
    """Stores provider authentication values."""

    def __init__(self, storage_type, credentials):
        self.type = storage_type
        self.creds = credentials

    def get_credential(self, key):
        """Return authentication credentials previously stored."""
        return self.creds.get(key, '')


class StorageConfig():
    """Parses providers authentication variables and defined outputs."""

    def __init__(self):
        # Create s3_auth with empty credentials
        self.s3_auth = {'default': AuthData('S3', None)}
        self.minio_auth = {}
        self.onedata_auth = {}
        self.webdav_auth = {}
        self.input = []
        self.output = []
        self._parse_config()

    @exception()
    def _parse_config(self):
        # Read output list
        output = ConfigUtils.read_cfg_var('output')
        if output != '':
            self.output = output
        else:
            get_logger().warning('There is no output defined for this function.')
        # Read input list
        input_value = ConfigUtils.read_cfg_var('input')
        if input_value != '':
            self.input = input_value
        else:
            get_logger().warning('There is no input defined for this function.')
        # Read storage_providers dict
        storage_providers = ConfigUtils.read_cfg_var('storage_providers')
        if (storage_providers and
                storage_providers != ''):
            # s3 storage provider auth
            if ('s3' in storage_providers
                    and storage_providers['s3']):
                self._validate_s3_creds(storage_providers['s3'])
            # minio storage provider auth
            if ('minio' in storage_providers
                    and storage_providers['minio']):
                self._validate_minio_creds(storage_providers['minio'])
            # onedata storage provider auth
            if ('onedata' in storage_providers
                    and storage_providers['onedata']):
                self._validate_onedata_creds(storage_providers['onedata'])
            # webdav storage providers auth
            if ('webdav' in storage_providers
                    and storage_providers['webdav']):
                self._validate_webdav_creds(storage_providers['webdav'])
        else:
            get_logger().warning('There is no storage provider defined for this function.')

    def _validate_minio_creds(self, minio_creds):
        if isinstance(minio_creds, dict):
            for provider_id in minio_creds:
                if ('access_key' in minio_creds[provider_id]
                        and minio_creds[provider_id]['access_key'] is not None
                        and minio_creds[provider_id]['access_key'] != ''
                        and 'secret_key' in minio_creds[provider_id]
                        and minio_creds[provider_id]['secret_key'] is not None
                        and minio_creds[provider_id]['secret_key'] != ''):
                    self.minio_auth[provider_id] = AuthData('MINIO', minio_creds[provider_id])
                else:
                    raise StorageAuthError(auth_type='MINIO')
        else:
            raise StorageAuthError(auth_type='MINIO')

    def _validate_s3_creds(self, s3_creds):
        if isinstance(s3_creds, dict):
            for provider_id in s3_creds:
                if ('access_key' in s3_creds[provider_id]
                        and s3_creds[provider_id]['access_key'] is not None
                        and s3_creds[provider_id]['access_key'] != ''
                        and 'secret_key' in s3_creds[provider_id]
                        and s3_creds[provider_id]['secret_key'] is not None
                        and s3_creds[provider_id]['secret_key'] != ''):
                    self.s3_auth[provider_id] = AuthData('S3', s3_creds[provider_id])
                else:
                    raise StorageAuthError(auth_type='S3')
        else:
            raise StorageAuthError(auth_type='S3')

    def _validate_onedata_creds(self, onedata_creds):
        if isinstance(onedata_creds, dict):
            for provider_id in onedata_creds:
                if ('oneprovider_host' in onedata_creds[provider_id]
                        and onedata_creds[provider_id]['oneprovider_host'] is not None
                        and onedata_creds[provider_id]['oneprovider_host'] != ''
                        and 'token' in onedata_creds[provider_id]
                        and onedata_creds[provider_id]['token'] is not None
                        and onedata_creds[provider_id]['token'] != ''
                        and 'space' in onedata_creds[provider_id]
                        and onedata_creds[provider_id]['space'] is not None
                        and onedata_creds[provider_id]['space'] != ''):
                    self.onedata_auth[provider_id] = AuthData('ONEDATA', onedata_creds[provider_id])
                else:
                    raise StorageAuthError(auth_type='ONEDATA')
        else:
            raise StorageAuthError(auth_type='ONEDATA')

    def _validate_webdav_creds(self, webdav_creds):
        if isinstance(webdav_creds, dict):
            for provider_id in webdav_creds:
                if ('hostname' in webdav_creds[provider_id]
                    and webdav_creds[provider_id]['hostname'] is not None
                    and webdav_creds[provider_id]['hostname'] != ''
                    and 'login' in webdav_creds[provider_id]
                    and webdav_creds[provider_id]['login'] is not None
                    and webdav_creds[provider_id]['login'] != ''
                    and 'password' in webdav_creds[provider_id]
                    and webdav_creds[provider_id]['password'] is not None
                    and webdav_creds[provider_id]['password'] != ''):
                    self.webdav_auth[provider_id] = AuthData('WEBDAV', webdav_creds[provider_id])
                else:
                    raise StorageAuthError(auth_type='WEBDAV')
        else:
            raise StorageAuthError(auth_type='WEBDAV')

    def _get_auth_data(self, storage_type, provider_id='default'):
        """Returns the authentication credentials by its type and id."""
        if storage_type == 'S3':
            return self.s3_auth.get(provider_id, None)
        elif storage_type == 'MINIO':
            return self.minio_auth.get(provider_id, None)
        elif storage_type == 'ONEDATA':
            return self.onedata_auth.get(provider_id, None)
        elif storage_type == 'WEBDAV':
            return self.webdav_auth.get(provider_id, None)
        return None

    def _get_input_auth_data(self, parsed_event):
        """Return the proper auth data from a storage_provider based on the event.

        This methods allows to filter ONEDATA provider when multiple inputs are defined."""
        storage_type = parsed_event.get_type()
        # Change storage type for dCache events to use WebDav
        if storage_type == 'DCACHE':
            storage_type = 'WEBDAV'
        if storage_type == 'ONEDATA':
            # Check input path and event object_key
            if hasattr(parsed_event, 'object_key'):
                # Get the onedata space from the event object_key
                event_space = parsed_event.object_key.strip('/').split('/', maxsplit=1)[0]
            for input_value in self.input:
                provider_type = StrUtils.get_storage_type(input_value.get('storage_provider'))
                if provider_type == storage_type:
                    provider_id = StrUtils.get_storage_id(input_value.get('storage_provider'))
                    if self.onedata_auth[provider_id].get_credential('space') == event_space:
                        return self._get_auth_data(storage_type, provider_id)
            raise StorageAuthError(auth_type='ONEDATA')
        elif storage_type == 'UNKNOWN':
            return self._get_auth_data(storage_type)
        else:
            return (self._get_auth_data(storage_type), self._get_auth_data(storage_type, parsed_event.provider_id))[parsed_event.provider_id != 'default']

    def download_input(self, parsed_event, input_dir_path):
        """Receives the event where the file information is and
        the tmp_dir_path where to store the downloaded file.

        Returns the file path where the file is downloaded."""
        auth_data = self._get_input_auth_data(parsed_event)
        stg_provider = create_provider(auth_data)
        get_logger().info('Found \'%s\' input provider', stg_provider.get_type())
        return stg_provider.download_file(parsed_event, input_dir_path)

    def upload_output(self, output_dir_path):
        """Receives the tmp_dir_path where the files to upload are stored and
        uploads files whose name matches the prefixes and suffixes specified
        in 'output'."""
        get_logger().info('Searching for files to upload in folder \'%s\'', output_dir_path)
        output_files = FileUtils.get_all_files_in_dir(output_dir_path)
        stg_providers = {}
        # Filter files by prefix and suffix
        for output in self.output:
            get_logger().info('Checking files for uploading to \'%s\' on path: \'%s\'',
                              output['storage_provider'],
                              output['path'])
            provider_type = StrUtils.get_storage_type(output['storage_provider'])
            provider_id = StrUtils.get_storage_id(output['storage_provider'])
            for file_path in output_files:
                file_name = file_path.replace(f'{output_dir_path}/', '')
                prefix_ok = False
                suffix_ok = False
                # Check prefixes
                if ('prefix' not in output
                        or len(output['prefix']) == 0):
                    prefix_ok = True
                else:
                    for pref in output['prefix']:
                        if file_name.startswith(pref):
                            prefix_ok = True
                            break
                if prefix_ok:
                    # Check suffixes
                    if ('suffix' not in output
                            or len(output['suffix']) == 0):
                        suffix_ok = True
                    else:
                        for suff in output['suffix']:
                            if file_name.endswith(suff):
                                suffix_ok = True
                                break
                    # Only upload file if name matches the prefixes and suffixes
                    if suffix_ok:
                        if provider_type not in stg_providers:
                            stg_providers[provider_type] = {}
                        if provider_id not in stg_providers[provider_type]:
                            auth_data = self._get_auth_data(provider_type, provider_id)
                            stg_providers[provider_type][provider_id] = create_provider(auth_data)
                        stg_providers[provider_type][provider_id].upload_file(file_path,
                                                                              file_name,
                                                                              output['path'])



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for the faassupervisor.storage module and classes."""

import unittest
from unittest import mock
from unittest.mock import call
from collections import namedtuple
from faassupervisor.storage.providers import get_bucket_name, get_file_key
from faassupervisor.storage.config import StorageConfig, AuthData, create_provider
from faassupervisor.storage.providers.local import Local
from faassupervisor.storage.providers.minio import Minio
from faassupervisor.storage.providers.onedata import Onedata
from faassupervisor.storage.providers.s3 import S3
from faassupervisor.events import parse_event
from faassupervisor.events.minio import MinioEvent
from faassupervisor.events.unknown import UnknownEvent
from faassupervisor.events.s3 import S3Event
from faassupervisor.events.onedata import OnedataEvent
from faassupervisor.utils import StrUtils

# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=protected-access

CONFIG_FILE_OK = """
name: test-func
input:
- storage_provider: onedata.test_onedata
  path: files
- storage_provider: onedata.test_onedata2
  path: files
output:
- storage_provider: s3
  path: bucket/folder
- storage_provider: minio.test_minio
  path: bucket
  suffix: ['txt', 'jpg']
  prefix: ['result-']
storage_providers:
  minio:
    test_minio:
        access_key: test_minio_access
        secret_key: test_minio_secret
  onedata:
    test_onedata:
        oneprovider_host: test_oneprovider.host
        token: test_onedata_token
        space: test_onedata_space
    test_onedata2:
        oneprovider_host: test_oneprovider.host
        token: test_onedata_token
        space: space_ok
"""

CONFIG_FILE_NO_OUTPUT = """
name: test-func
storage_providers:
  minio:
    test_minio:
        access_key: test_minio_access
        secret_key: test_minio_secret
"""

CONFIG_FILE_NO_STORAGE_PROVIDER = """
name: test-func
output:
- storage_provider: s3
  path: bucket/folder
"""

CONFIG_FILE_INVALID_MINIO = """
name: test-func
storage_providers:
  minio:
    secret_key: test_minio_secret
"""

CONFIG_FILE_INVALID_ONEDATA = """
name: test-func
storage_providers:
  onedata:
    token: test_onedata_token
    space: test_onedata_space
"""

CONFIG_FILE_INVALID_S3 = """
name: test-func
storage_providers:
  s3:
    user: test
"""

ONEDATA_EVENT = """
{
    "Key": "/space_ok/files/file.txt",
    "Records": [
        {
            "objectKey": "file.txt",
            "objectId": "1234",
            "eventTime": "2019-02-07T09:51:04.347823",
            "eventSource": "OneTrigger"
        }
    ]
}
"""

class AuthDataTest(unittest.TestCase):

    def test_create_auth_data(self):
        auth = AuthData('MINIO', {'access_key': 'test'})
        self.assertEqual(auth.type, 'MINIO')
        self.assertEqual(auth.creds, {'access_key': 'test'})

    def test_get_auth_data_credential(self):
        auth = AuthData('MINIO', {'access_key': 'test'})
        self.assertEqual(auth.get_credential('access_key'), 'test')
        self.assertEqual(auth.get_credential('k'), '')


class StorageConfigTest(unittest.TestCase):

    def test_parse_config_valid(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            config = StorageConfig()
            expected_output = [
                {
                    'storage_provider': 's3',
                    'path': 'bucket/folder'
                }, {
                    'storage_provider': 'minio.test_minio',
                    'path': 'bucket',
                    'suffix': ['txt', 'jpg'],
                    'prefix': ['result-']
                }
            ]
            self.assertEqual(config.output, expected_output)
            self.assertEqual(config.minio_auth['test_minio'].type, 'MINIO')
            self.assertEqual(config.minio_auth['test_minio'].get_credential('access_key'), 'test_minio_access')
            self.assertEqual(config.minio_auth['test_minio'].get_credential('secret_key'), 'test_minio_secret')
            self.assertEqual(config.onedata_auth['test_onedata'].type, 'ONEDATA')
            self.assertEqual(config.onedata_auth['test_onedata'].get_credential('oneprovider_host'), 'test_oneprovider.host')
            self.assertEqual(config.onedata_auth['test_onedata'].get_credential('token'), 'test_onedata_token')
            self.assertEqual(config.onedata_auth['test_onedata'].get_credential('space'), 'test_onedata_space')
            self.assertEqual(config.s3_auth['default'].type, 'S3')

    def test_parse_config_no_output(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_NO_OUTPUT)},
                             clear=True):
            StorageConfig()
            self.assertLogs('There is no output defined for this function.',
                             level='WARNING')

    def test_parse_config_no_storage_provider(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_NO_STORAGE_PROVIDER)},
                             clear=True):
            StorageConfig()
            self.assertLogs('There is no storage provider defined for this function.',
                             level='WARNING')

    def test_parse_config_invalid_minio(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_MINIO)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'MINIO\' is not well-defined.',
                                level='ERROR')

    def test_parse_config_invalid_s3(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_S3)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'S3\' is not well-defined.',
                                level='ERROR')

    def test_parse_config_invalid_onedata(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_INVALID_ONEDATA)},
                             clear=True):
            with self.assertRaises(SystemExit):
                StorageConfig()
                self.assertLogs('The storage authentication of \'ONEDATA\' is not well-defined.',
                                level='ERROR')

    def test_get_minio_auth(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            minio_auth = StorageConfig()._get_auth_data('MINIO', 'test_minio')
            self.assertEqual(minio_auth.type, 'MINIO')
            self.assertEqual(minio_auth.get_credential('access_key'),
                             'test_minio_access')
            self.assertEqual(minio_auth.get_credential('secret_key'),
                             'test_minio_secret')

    def test_get_s3_auth(self):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            s3_auth = StorageConfig()._get_auth_data('S3')
            self.assertEqual(s3_auth.type, 'S3')

    def test_get_onedata_auth(self):
        # TODO: check _get_input_auth_data() with custom Onedata event
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            onedata_auth = StorageConfig()._get_auth_data('ONEDATA', 'test_onedata')
            self.assertEqual(onedata_auth.type, 'ONEDATA')
            self.assertEqual(onedata_auth.get_credential('oneprovider_host'),
                             'test_oneprovider.host')
            self.assertEqual(onedata_auth.get_credential('token'),
                             'test_onedata_token')
            self.assertEqual(onedata_auth.get_credential('space'),
                             'test_onedata_space')
            # Test _get_input_auth_data() for getting the proper auth data from an event
            parsed_event = parse_event(ONEDATA_EVENT)
            onedata2_auth = StorageConfig()._get_input_auth_data(parsed_event)
            self.assertEqual(onedata2_auth.get_credential('space'), 'space_ok')

    def test_get_invalid_auth(self):
        invalid_auth = StorageConfig()._get_auth_data('INVALID_TYPE')
        self.assertIsNone(invalid_auth)

    @mock.patch('faassupervisor.storage.providers.local.Local.download_file')
    def test_download_input(self, mock_download_file):
        event = UnknownEvent('')
        StorageConfig().download_input(event, '/tmp/test')
        mock_download_file.assert_called_once_with(event, '/tmp/test')

    @mock.patch('faassupervisor.utils.FileUtils.get_all_files_in_dir')
    @mock.patch('faassupervisor.storage.providers.s3.S3.upload_file')
    @mock.patch('faassupervisor.storage.providers.minio.Minio.upload_file')
    def test_upload_output(self, mock_minio, mock_s3, mock_get_files):
        with mock.patch.dict('os.environ',
                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
                             clear=True):
            files = [
                '/tmp/test/file1.txt',
                '/tmp/test/file1.jpg',
                '/tmp/test/result-file.txt',
                '/tmp/test/result-file.out',
                '/tmp/test/file2.txt',
                '/tmp/test/file2.out'
            ]
            mock_get_files.return_value = files
            StorageConfig().upload_output('/tmp/test')
            self.assertEqual(mock_minio.call_count, 1)
            self.assertEqual(mock_minio.call_args,
                             call('/tmp/test/result-file.txt',
                                  'result-file.txt',
                                  'bucket'))
            self.assertEqual(mock_s3.call_count, 6)
            for i, f in enumerate(files):
                self.assertEqual(mock_s3.call_args_list[i],
                                 call(f, f.split('/')[3], 'bucket/folder'))

#    def test_upload_real_output(self):
#        with mock.patch.dict('os.environ',
#                             {'FUNCTION_CONFIG': StrUtils.utf8_to_base64_string(CONFIG_FILE_OK)},
#                             clear=True):
#            files = [
#                '/home/caterina/Documentos/test/y1',
#                '/home/caterina/Documentos/test/y2',
#            ]
#            StorageConfig().upload_output('/home/caterina/Documentos/test')


class ProvidersModuleTest(unittest.TestCase):

    def test_get_bucket_name(self):
        self.assertEqual(get_bucket_name('bucket/folder1/folder2'), 'bucket')
        self.assertEqual(get_bucket_name('bucket'), 'bucket')

    def test_get_file_key(self):
        self.assertEqual(get_file_key('bucket/folder1/folder2', 'file'),
                         'folder1/folder2/file')
        self.assertEqual(get_file_key('bucket', 'file'), 'file')


class LocalProviderTest(unittest.TestCase):

    def test_create_local_provider(self):
        provider = create_provider(None)
        self.assertEqual(provider.get_type(), 'LOCAL')

    def test_download_file(self):
        provider = Local(None)
        parsed_event = mock.Mock(spec=UnknownEvent)
        provider.download_file(parsed_event, '/tmp/local')
        parsed_event.save_event.assert_called_once_with('/tmp/local')


class MinioProviderTest(unittest.TestCase):

    MINIO_CREDS = {
            'access_key': 'test_minio_access',
            'secret_key': 'test_minio_secret'
    }

    def test_create_minio_provider(self):
        minio_auth = AuthData('MINIO', self.MINIO_CREDS)
        provider = create_provider(minio_auth)
        self.assertEqual(provider.get_type(), 'MINIO')
        self.assertEqual(provider.stg_auth.creds, self.MINIO_CREDS)

    @mock.patch('boto3.client')
    def test_get_client_default_endpoint(self, mock_boto):
        Minio(AuthData('MINIO', self.MINIO_CREDS))
        mock_boto.assert_called_once_with('s3',
                                           endpoint_url='http://minio-service.minio:9000',
                                           region_name=None,
                                           verify=True,
                                           aws_access_key_id='test_minio_access',
                                           aws_secret_access_key='test_minio_secret')

    @mock.patch('boto3.client')
    def test_get_client_custom_endpoint(self, mock_boto):
        Minio(AuthData('MINIO', {**self.MINIO_CREDS,
                                 'endpoint': 'https://test.endpoint'}))
        mock_boto.assert_called_once_with('s3',
                                           endpoint_url='https://test.endpoint',
                                           region_name=None,
                                           verify=True,
                                           aws_access_key_id='test_minio_access',
                                           aws_secret_access_key='test_minio_secret')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        minio_provider = Minio(AuthData('MINIO', self.MINIO_CREDS))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Create mock event and event properties
            event = mock.Mock(spec=MinioEvent)
            type(event).bucket_name = mock.PropertyMock(return_value='minio_bucket')
            type(event).file_name = mock.PropertyMock(return_value='minio_file')
            type(event).object_key = mock.PropertyMock(return_value='minio_file')
            file_path = minio_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/minio_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/minio_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('minio_bucket',
                                                     'minio_file',
                                                     mopen.return_value))

    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto):
        minio_provider = Minio(AuthData('MINIO', self.MINIO_CREDS))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            minio_provider.upload_file('/tmp/output/processed.jpg', 'processed.jpg', 'minio_bucket')
            # Check file reading call
            mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
            # Check boto client upload call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().upload_fileobj(mopen.return_value,
                                                   'minio_bucket',
                                                   'processed.jpg'))


class OnedataProviderTest(unittest.TestCase):

    ONEDATA_CREDS = {
        'oneprovider_host': 'test_oneprovider.host',
        'token': 'test_onedata_token',
        'space': 'test_onedata_space'
    }
    def test_create_onedata_provider(self):
        onedata_auth = AuthData('ONEDATA', self.ONEDATA_CREDS)
        provider = create_provider(onedata_auth)
        self.assertEqual(provider.get_type(), 'ONEDATA')
        self.assertEqual(provider.stg_auth.creds, self.ONEDATA_CREDS)

    @mock.patch('requests.get')
    def test_download_file(self, mock_requests):
        onedata_provider = Onedata(AuthData('ONEDATA', self.ONEDATA_CREDS))
        # Mock requests.get response
        response = namedtuple('response', ['content', 'status_code'])
        mock_requests.return_value = response('test response', 200)
        # Create mock event
        event = mock.Mock(spec=OnedataEvent)
        type(event).file_name = mock.PropertyMock(return_value='onedata_file')
        type(event).object_key = mock.PropertyMock(return_value='/onedata_file_key')
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            file_path = onedata_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/onedata_file')
            # Check request to onedata endpoint
            mock_requests.assert_called_once_with('https://test_oneprovider.host/cdmi/onedata_file_key',
                                                  headers={'X-Auth-Token': 'test_onedata_token'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/input/onedata_file', 'wb')

    @mock.patch('requests.get')
    @mock.patch('requests.put')
    def test_upload_file(self, mock_put, mock_get):
        onedata_provider = Onedata(AuthData('ONEDATA', self.ONEDATA_CREDS))
        response = namedtuple('response', ['status_code'])
        # Mock requests.get response (check if folder exists)
        mock_get.return_value = response(200)
        # Mock requests.put response
        mock_put.return_value = response(202)
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            onedata_provider.upload_file('/tmp/output/onedata_file', 'onedata_file', 'onedata_path')
            # Check request to onedata endpoint
            mock_get.assert_called_once_with(
                'https://test_oneprovider.host/cdmi/test_onedata_space/onedata_path/',
                headers={**onedata_provider._CDMI_VERSION_HEADER,
                         'X-Auth-Token': 'test_onedata_token'})
            mock_put.assert_called_once_with(
                'https://test_oneprovider.host/cdmi/test_onedata_space/onedata_path/onedata_file',
                data=mopen.return_value,
                headers={'X-Auth-Token': 'test_onedata_token'})
            # Check file writing
            mopen.assert_called_once_with('/tmp/output/onedata_file', 'rb')


class S3ProviderTest(unittest.TestCase):

    S3_CREDS = {
        'access_key': 'test_s3_access',
        'secret_key': 'test_s3_secret'
    }

    def test_create_s3_provider(self):
        minio_auth = AuthData('S3', None)
        provider = create_provider(minio_auth)
        self.assertEqual(provider.get_type(), 'S3')
        self.assertIsNone(provider.stg_auth.creds)

    @mock.patch('boto3.client')
    def test_get_client_without_creds(self, mock_boto):
        S3(AuthData('S3', None))
        mock_boto.assert_called_once_with('s3')

    @mock.patch('boto3.client')
    def test_get_client_with_creds(self, mock_boto):
        S3(AuthData('S3', self.S3_CREDS))
        mock_boto.assert_called_once_with('s3',
                                           region_name=None,
                                           aws_access_key_id='test_s3_access',
                                           aws_secret_access_key='test_s3_secret')

    @mock.patch('boto3.client')
    def test_download_file(self, mock_boto):
        s3_provider = S3(AuthData('S3', None))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            # Create mock event and event properties
            event = mock.Mock(spec=S3Event)
            type(event).bucket_name = mock.PropertyMock(return_value='s3_bucket')
            type(event).file_name = mock.PropertyMock(return_value='s3_file')
            type(event).object_key = mock.PropertyMock(return_value='s3_folder/s3_file')
            file_path = s3_provider.download_file(event, '/tmp/input')
            # Check returned file path
            self.assertEqual(file_path, '/tmp/input/s3_file')
            # Check file writing call
            mopen.assert_called_once_with('/tmp/input/s3_file', 'wb')
            # Check boto client download call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().download_fileobj('s3_bucket',
                                                     's3_folder/s3_file',
                                                     mopen.return_value))

    @mock.patch('boto3.client')
    def test_upload_file(self, mock_boto):
        s3_provider = S3(AuthData('S3', None))
        # Mock file management
        mopen = mock.mock_open()
        with mock.patch('builtins.open', mopen, create=True):
            s3_provider.upload_file('/tmp/output/processed.jpg', 'processed.jpg', 's3_bucket/s3_folder')
            # Check file reading call
            mopen.assert_called_once_with('/tmp/output/processed.jpg', 'rb')
            # Check boto client upload call
            self.assertEqual(mock_boto.mock_calls[1],
                             call().upload_fileobj(mopen.return_value,
                                                   's3_bucket',
                                                   's3_folder/processed.jpg'))
